package se.su.dsv.scipro.util;

import java.util.function.Consumer;
import java.util.function.Function;

public abstract class Either<E, A> {
    public static <E, A> Either<E, A> left(E error) {
        return new Left<>(error);
    }
    public static <E, A> Either<E, A> right(A a) {
        return new Right<>(a);
    }

    public abstract <X> X fold(Function<E, X> ifError, Function<? super A, X> f);

    public <B> Either<E, B> flatMap(Function<? super A, Either<E, B>> f) {
        return fold(Either::left, f);
    }

    public <B> Either<E, B> map(Function<? super A, B> f) {
        return flatMap(f.andThen(Either::right));
    }

    public boolean isLeft() {
        return fold(e -> true, a -> false);
    }

    public boolean isRight() {
        return fold(e -> false, a -> true);
    }

    public void foreach(Consumer<? super A> sideEffect) {
        fold(e -> null, a -> { sideEffect.accept(a); return null; });
    }

    @Override
    public String toString() {
        return fold(
                error -> "Left(" + error.toString() + ")",
                a -> "Right(" + a.toString() + ")"
        );
    }

    private static class Right<E, A> extends Either<E, A> {
        private final A a;

        Right(final A a) {
            this.a = a;
        }

        @Override
        public <X> X fold(final Function<E, X> ifError, final Function<? super A, X> f) {
            return f.apply(a);
        }
    }

    private static class Left<E, A> extends Either<E, A> {
        private final E error;

        public Left(final E error) {
            this.error = error;
        }

        @Override
        public <X> X fold(final Function<E, X> ifError, final Function<? super A, X> f) {
            return ifError.apply(error);
        }
    }
}
